from kratos import Generator, TestBench, initial, assert_, delay, Sequence, verilog, always, posedge


def tb_dut_setup():
    dut = Generator("mod")
    dut.wire(dut.output("out", 1), dut.input("in", 1))
    dut.wire(dut.var("val", 1), dut.ports["in"])

    tb = TestBench()

    tb.add_child("dut", dut)
    in_ = tb.var("in", 1)
    out_ = tb.var("out", 1)
    tb.wire(dut.ports["in"], in_)
    tb.wire(out_, dut.ports["out"])
    return dut, tb


def test_tb_codegen(check_gold):
    dut, tb = tb_dut_setup()

    @initial
    def code():
        tb.vars["in"] = 1
        assert_(tb.vars.out == 1)
        # access internal signal
        assert_(dut.vars.val == 1)

    tb.add_always(code)

    check_gold(tb, "test_tb_codegen")


def test_tb_delay(check_gold):
    dut, tb = tb_dut_setup()

    @initial
    def code():
        delay(1, tb.vars["in"].assign(1))

    tb.add_always(code)
    check_gold(tb, "test_tb_delay")


def test_tb_sequence(check_gold):
    from kratos.util import clock
    from kratos import PropertyAction

    dut, tb = tb_dut_setup()
    # add a clock and wire them together
    tb.wire(dut.clock("clk"), clock(tb.var("clk", 1)))

    seq = Sequence(tb.vars["in"] == 1)
    seq.imply(tb.vars.out == 1).wait(1).imply(tb.vars.out == 0)

    p = tb.property("test_out", seq)
    p.action = PropertyAction.Assert

    check_gold(tb, "test_tb_sequence")


def test_display_stmt():
    from kratos.util import display
    mod = Generator("gen")
    a = mod.var("a", 1)

    @initial
    def code():
        display("%d", a)

    mod.add_always(code)

    src = verilog(mod)["gen"]
    assert '$display ("%d", a);\n' in src

    import _kratos
    from kratos.passes import check_non_synthesizable_content
    try:
        check_non_synthesizable_content(mod.internal_generator)
        assert False
    except _kratos.exception.UserException:
        pass


def test_event_control_stmt():
    mod = Generator("gen")
    a = mod.var("a", 1)

    @initial
    def code():
        posedge(a)
        delay(1, None)
        a = 1

    mod.add_always(code)

    src = verilog(mod)["gen"]
    assert "#1;\n" in src
    assert "\n  @(posedge a);\n" in src


def test_always():
    mod = Generator("mod")
    clk = mod.var("clk", 1)

    @always
    def code():
        delay(5, clk.assign(~clk))

    mod.add_always(code)

    src = verilog(mod)["mod"]
    assert "always begin\n" in src
    assert "#5 clk = ~clk;\n" in src


def test_file_ops(check_gold):
    from kratos.util import fopen, fscanf, fclose
    mod = Generator("mod")
    a = mod.var("a", 1)
    b = mod.var("b", 1)
    c = mod.var("c", 32)
    fd = mod.var("fd", 32)

    @initial
    def code():
        fd = fopen("test.sv", "r")
        for i in range(10):
            c = fscanf(fd, "%d %d", a, b)
        fclose(fd)

    mod.add_always(code)
    check_gold(mod, "test_file_ops")


def test_finish():
    from kratos.util import finish
    mod = Generator("mod")

    @initial
    def code():
        finish(0)
        finish()

    mod.add_code(code)
    src = verilog(mod)["mod"]
    assert "$finish (32'h0);\n" in src
    assert "$finish ();\n" in src


if __name__ == "__main__":
    from conftest import check_gold_fn
    test_file_ops(check_gold_fn)



import _kratos
from .generator import Generator
from _kratos import get_fn_ln


def assert_(expr):
    if isinstance(expr, _kratos.Var):
        stmt = _kratos.AssertValueStmt(expr)
    else:
        assert isinstance(expr, _kratos.Property)
        expr.action = _kratos.PropertyAction.Assert
        stmt = _kratos.AssertPropertyStmt(expr)
    if expr.generator.debug:
        stmt.add_fn_ln(get_fn_ln())
    return stmt


def assume(prop):
    prop.action = _kratos.PropertyAction.Assume
    stmt = _kratos.AssertPropertyStmt(prop)
    return stmt


def cover(prop):
    prop.action = _kratos.PropertyAction.Cover
    stmt = _kratos.AssertPropertyStmt(prop)
    return stmt


def delay(num, stmt, lhs=True):
    event = _kratos.EventControl(num)
    if stmt is None:
        stmt = _kratos.EventDelayStmt(event)
    else:
        assert isinstance(stmt, _kratos.AssignStmt)
        event.delay_side = _kratos.DelaySide.Left if lhs else _kratos.DelaySide.Right
        stmt.delay = event
    return stmt


class TestBench(Generator):
    # disable pytest collection
    __test__ = False

    def __init__(self, top_name: str = "TOP"):
        gen = Generator.get_context().testbench(top_name)
        super(TestBench, self).__init__(top_name, internal_generator=gen)



from kratos import Generator, always_ff, verilog, posedge, const


def setup_fsm(fsm, out_, in_):
    # add outputs
    fsm.output(out_)
    # add states
    red = fsm.add_state("Red")
    blue = fsm.add_state("Blue")
    # set the state transition
    red.next(red, in_ == 0)
    red.next(blue, in_ == 1)
    blue.next(red, in_ == 1)
    # fill in outputs
    red.output(out_, 2)
    blue.output(out_, 1)
    # set the start case
    fsm.set_start_state("Red")


def test_fsm(check_gold, check_file):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color")
    # setup FSM
    setup_fsm(fsm, out_, in_)

    check_gold(mod, "test_fsm", optimize_if=False)
    # output fsm graph
    dot = fsm.dot_graph()
    check_file(dot, "test_fsm.dot")
    csv = fsm.output_table()
    check_file(csv, "test_fsm.csv")


def test_fsm_mealy(check_gold):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color")
    # setup FSM
    setup_fsm(fsm, out_, in_)
    # use mealy
    fsm.is_moore = False
    check_gold(mod, "test_fsm_mealy", optimize_if=False)


def test_fsm_default_output():
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    fsm = mod.add_fsm("Color")
    fsm.output(out_, const(2, 2))
    # add states
    red = fsm.add_state("Red")
    blue = fsm.add_state("Blue")
    red.next(blue, in_ == 0)
    blue.next(red, in_ == 1)
    blue.output(out_, 1)
    fsm.set_start_state("Red")
    src = verilog(mod)["mod"]
    assert "out = 2'h2;" in src


def test_fsm_state(check_gold):
    from kratos import negedge
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    clk = mod.clock("clk")
    rst = mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color", reset_high=False)
    # setup FSM
    setup_fsm(fsm, out_, in_)
    # realize fsm now
    fsm.realize()
    current_state = fsm.current_state
    # create an enum var based on current_state
    state_enum = current_state.enum_type()
    # can create a variable from the enum definition
    # notice that variable s will be optimized away if not used
    s = mod.enum_var("s", state_enum)
    c = mod.var("counter", 1)

    @always_ff((posedge, clk), (negedge, rst))
    def counter():
        if rst.r_not():
            c = 0
        elif current_state == state_enum.Red:
            c = c + 1

    mod.add_always(counter)
    check_gold(mod, "test_fsm_state", optimize_if=False)


def test_nested_fsm(check_gold, check_file):
    mod = Generator("mod", debug=True)
    out_ = mod.output("out", 2)
    in_ = mod.input("in", 2)
    # fsm requires a clk and async rst
    mod.clock("clk")
    mod.reset("rst")
    # add a dummy fsm
    fsm = mod.add_fsm("Color")
    setup_fsm(fsm, out_, in_)
    second_fsm = mod.add_fsm("HSV")
    fsm.add_child_fsm(second_fsm)
    idle = second_fsm.add_state("idle")
    idle.next(fsm["Red"], in_ == 0)
    fsm["Red"].next(idle, in_ == 2)
    second_fsm.output(out_)
    idle.output(out_, 2)

    dot = fsm.dot_graph()
    check_file(dot, "test_nested_fsm.dot")
    check_gold(mod, "test_nested_fsm", optimize_if=False)



from kratos import Generator, Attribute


def test_wrapper_flatten_generator(check_gold):
    mod = Generator("mod")
    a = mod.input("a", 4, size=[3, 2])
    b = mod.output("b", 4, size=[3, 2])
    attr = Attribute()
    attr.value_str = "a"
    a.add_attribute(attr)
    mod.wire(a, b)
    from _kratos import create_wrapper_flatten
    wrapper = create_wrapper_flatten(mod.internal_generator, "wrapper")
    wrapper = Generator(wrapper.name, internal_generator=wrapper)
    check_gold(wrapper, gold_name="test_wrapper_flatten_generator",
               optimize_passthrough=False)
    a_0_0 = wrapper.ports.a_0_0
    attrs = a_0_0.find_attribute(lambda x: x.value_str == "a")
    assert len(attrs) == 1



from kratos import Generator, PortDirection, PortType, always_ff, \
    verilog, StmtException, posedge


def test_illegal_assignment_blocking():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod1", True)
            self.in_ = self.port("in", 1, PortDirection.In)
            self.out_ = self.port("out", 1, PortDirection.Out)
            self.clk_ = self.port("clk", 1, PortDirection.In, PortType.Clock)

            self.wire(self.out_, self.in_)

            self.add_always(self.code)

        @always_ff((posedge, "clk"))
        def code(self):
            self.out_ = 1

    try:
        mod = Mod()
        verilog(mod)
        assert False
    except StmtException as ex:
        print(ex)
        assert True


def test_async_no_latch():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)
            clk = self.clock("clk")
            rst = self.reset("rst")
            cen = self.input("cen", 1, PortType.ClockEnable)
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)

            @always_ff((posedge, "clk"), (posedge, "rst"))
            def code():
                if rst:
                    out_ = 0
                elif cen:
                    out_ = in_

            self.add_always(code)

    mod = Mod()
    verilog(mod)


def test_async_latch():
    class Mod(Generator):
        def __init__(self):
            super().__init__("mod", True)
            clk = self.clock("clk")
            rst = self.reset("rst")
            cen = self.input("cen", 1, PortType.ClockEnable)
            in_ = self.input("in", 1)
            out_ = self.output("out", 1)

            @always_ff((posedge, "clk"), (posedge, "rst"))
            def code():
                if rst:
                    out_ = 0

            self.add_always(code)

    mod = Mod()
    try:
        verilog(mod)
        assert False
    except StmtException:
        assert True


def test_always_latch(check_gold):
    from kratos import always_latch
    mod = Generator("mod")
    a = mod.input("a", 1)
    b = mod.output("b", 1)

    @always_latch
    def code():
        if a:
            b = 1

    mod.add_always(code)
    check_gold(mod, gold_name="test_always_latch", reorder_stmts=True)


def test_generator_port_connected():
    child = Generator("child")
    parent = Generator("parent")
    in1 = parent.var("a", 1)
    in2 = child.input("a", 1)
    out1 = parent.output("b", 1)
    out2 = child.output("b", 1)
    un_c = child.input("c", 1)
    parent.add_child("inst", child, a=in1, b=out1)
    assert in2.connected()
    assert out2.connected()
    assert not un_c.connected()


def test_register_extraction():
    mod = Generator("mod")
    a = mod.input("a", 4)
    b = mod.output("b", 4)
    rst = mod.reset("rst")
    clk = mod.clock("clk")

    @always_ff((posedge, "clk"), (posedge, "rst"))
    def code():
        if rst:
            b = 0
        else:
            b = a + b

    mod.add_always(code)

    import _kratos
    regs = _kratos.passes.extract_register_names(mod.internal_generator)
    assert len(regs) == 1
    assert "mod.b" in regs


def test_merge_const_port_assignment():
    mod = Generator("mod")
    a = mod.var("a", 1)
    child = Generator("child")
    b = child.input("b", 1)
    mod.add_child("child", child, b=a)
    mod.add_stmt(a.assign(1))
    v = verilog(mod)["mod"]
    assert ".b(1'h1)" in v



import kratos
from kratos import Generator, signed, unsigned, verilog
from kratos.util import reduce_or, Const, const, countones
import _kratos.exception


def test_expr():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = mod.var("b", 2)
    expr = a + b
    assert str(expr) == "a + b"


def test_slice():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = a[0]
    assert b.width == 1
    b = a[1, 0]
    assert b.width == 2
    assert str(b) == "a[1:0]"
    p = mod.param("p", value=32, initial_value=32)
    c = mod.input("c", p)
    d = c[19, 0]
    assert d.width == 20


def test_assign():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = mod.var("b", 2)
    assign = a.assign(b)
    assert id(assign.left) == id(a)
    assert id(assign.right) == id(b)


def test_signed():
    mod = Generator("mod")
    a = mod.var("a", 2)
    c = signed(a)
    assert str(c) == "signed'(a)"


def test_unsigned():
    mod = Generator("mod")
    a = mod.var("a", 2, is_signed=True)
    c = unsigned(a)
    assert str(c) == "unsigned'(a)"


def test_reduce_or():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = mod.var("b", 2)
    expr = reduce_or(a, b)
    assert str(expr) == "a | b"
    assert str(reduce_or(a)) == "a"


def test_const():
    a = Const[4](2)
    b = const(2, 4)
    assert str(a) == str(b)
    assert str(a) == "4'h2"


def test_width():
    mod = Generator("mod", True)
    a = mod.var("a", 1)
    a.width = 2
    file_names = list(a.fn_name_ln)
    assert len(file_names) == 2
    assert file_names[0][0] == file_names[1][0]
    assert abs(file_names[0][1] - file_names[1][1]) == 1


def test_slice_same():
    mod = Generator("mod")
    a = mod.var("a", 2, size=2)
    b = a[1, 1]
    assert str(b) == "a[1]"


def test_explicit_array():
    mod = Generator("mod")
    a = mod.var("a", 2, explicit_array=True, packed=True)
    b = a[0, 0]
    assert b.width == 2
    c = mod.var("c", 2, explicit_array=True, packed=True)
    d = mod.var("d", 1)
    e = mod.var("e", 2)
    mod.wire(c, a)
    mod.wire(e, a[d])
    src = verilog(mod)["mod"]
    assert "logic [0:0][1:0] a" in src


def test_neq():
    mod = Generator("mod")
    a = mod.var("a", 2)
    b = a != 2
    assert str(b) == "a != 2'h2"


def test_nested_ternary():
    from kratos import mux
    mod = Generator("mod")
    a = mod.var("a", 1)
    b = mod.var("b", 1)
    c = a + mux(a, a, b)
    assert str(c) == "a + (a ? a: b)"


def test_packed_unpacked():
    import _kratos
    mod = Generator("mod")
    a = mod.var("a", 2, size=2, packed=True)
    b = mod.var("b", 2, size=2)
    c = const(2, 4)
    d = mod.var("d", 4, explicit_array=True)
    try:
        a.assign(b)
        assert False
    except _kratos.exception.StmtException:
        assert True
    a.assign(4)
    try:
        b.assign(c)
        assert False
    except _kratos.exception.StmtException:
        assert True
    try:
        d.assign(a)
        assert False
    except _kratos.exception.StmtException:
        assert True

    # test for slices
    e = mod.var("e", 2, size=2, packed=True)
    e_0 = e[0]
    assert e_0.is_packed


def test_md_array():
    mod = Generator("mod")
    a = mod.var("a", 16, size=(4, 2), packed=True)
    b = mod.var("b", 16, size=(4, 2), packed=True)
    c = mod.var("c", 16, size=(2,), packed=True)
    mod.wire(b, a)
    mod.wire(c, a[1])
    src = verilog(mod)["mod"]
    assert "logic [3:0][1:0][15:0] a;" in src


def test_enum_cast():
    from kratos import enum
    e = enum("Type", ["a", "b"])
    mod = Generator("mod")
    var = mod.enum_var("v", e)
    casted = enum(const(1, 1), e)
    assert str(casted) == "Type'(1'h1)"
    var.assign(casted)


def test_bool_except():
    mod = Generator("mod")
    a = mod.var("a", 1)
    try:
        _ = not a
        assert False
    except _kratos.exception.InvalidConversionException:
        assert True


def test_single_concat():
    from kratos import concat
    mod = Generator("mod")
    a = mod.var("a", 1)
    r = concat([a])
    assert id(r) == id(a)


def test_resize():
    from kratos import resize
    mod = Generator("mod")
    a = mod.var("a", 1)
    b = resize(a, 16)
    assert b.width == 16
    assert str(b) == "16'(a)"


def test_power():
    mod = Generator("mod")
    a = mod.var("a", 5)
    b = 2 ** a
    assert str(b) == "5'h2 ** a"


def test_lshift():
    mod = Generator("mod")
    a = mod.var("a", 5)
    b = mod.var("b", 5)
    c = a << b
    assert str(c) == "a << b"


def test_big_num():
    from kratos import const
    a = const(2 ** 100, 120)
    assert a.is_bignum
    s = str(a)
    assert s == "120'h10000000000000000000000000"
    b = const(1, 120)
    assert not b.is_bignum


def test_var_slice_param():
    mod = Generator("mod")
    a = mod.parameter("a", initial_value=10)
    b = mod.var("b", a, size=4)
    c = b[0]
    assert c.width == 10
    a.value = 12
    assert b.width == 12
    assert c.width == 12


def test_concat_assign():
    from kratos import concat
    from _kratos.exception import VarException
    mod = Generator("mod")
    a = mod.var("a", 4)
    b = mod.var("b", 4)
    c = mod.var("c", 8)
    concat(a, b).assign(c)
    try:
        (a + b).assign(c)
        assert False
    except VarException:
        pass


def test_resize_const():
    from kratos import resize
    a = resize(0, 2)
    assert str(a) == "2'h0"
    b = resize(2, 4)
    assert str(b) == "4'h2"
    c = resize(-1, 2)
    assert str(c) == "-2'h1"


def test_bit_task():
    mod = Generator("mod")
    a = mod.var("a", 10)
    b = countones(a)
    assert str(b) == "$countones (a)"


def test_slice_signed():
    from kratos import signed
    mod = Generator("mod")
    a = mod.var("a", width=15, size=5, explicit_array=True, is_signed=True)
    assert a.signed
    assert not a[0].signed
    s = signed(a[1])
    assert str(s) == "signed'(a[1])"


def test_infer_const():
    mod = Generator("mod")
    a = mod.var("a", 4)
    b = mod.var("b", 1)
    expr = a == 1
    assert str(expr) == "a == 4'h1"
    expr = kratos.mux(b, 0, a)
    assert str(expr) == "b ? 4'h0: a"


def test_duplicate():
    mod = Generator("mod")
    a = mod.var("a", 4)
    b = mod.param("p", value=4)
    c = a.duplicate(2)
    d = a.duplicate(b)
    assert str(c) == "{32'h2{a}}"
    assert str(d) == "{p{a}}"


def test_string_const():
    mod = Generator("mod")
    s = const("kratos", 8 * 8)
    assert str(s) == '"kratos"'
    a = mod.var("a", 64)
    mod.wire(a, s)


def test_display_call():
    from kratos.util import display
    mod = Generator("mod")
    c = display("test", mod.var("a", 1))
    assert str(c) == '$display ("test", a)'


if __name__ == "__main__":
    test_display_call()



from kratos import Generator, always_ff, verilog, Attribute, posedge


class PassThroughMod(Generator):
    def __init__(self, is_clone: bool = False):
        super().__init__("mod1", True, is_clone)
        self.in_ = self.input("in", 1)
        self.out_ = self.output("out", 1)
        self.wire(self.out_, self.in_)


class PassThroughTop(Generator):
    def __init__(self):
        super().__init__("top", True)

        self.input("in", 1)
        self.output("out", 1)

        pass_through = PassThroughMod()
        self.add_child("pass", pass_through)
        self.wire(self["pass"].ports["in"], self.ports["in"], )

        self.wire(self.ports.out, self["pass"].ports.out)


def test_wire_merge(check_gold):
    class TestModule(Generator):
        def __init__(self, width):
            super().__init__("Test")
            self.input("in", width)
            self.output("out", width)

            for i in range(width):
                self.wire(self.ports["out"][i], self.ports["in"][i])

    mod = TestModule(4)
    check_gold(mod, "test_wire_merge")


def test_remove_child():
    top = PassThroughTop()
    child = top["pass"]
    assert child in top
    top.remove_child_generator(child)
    assert child not in top
    # top should be empty now
    assert top.stmts_count == 0


def test_simple_pipeline(check_gold):
    mod = PassThroughMod()
    # add a clock
    mod.clock("clk")
    attr = Attribute()
    attr.type_str = "pipeline"
    attr.value_str = "2"
    mod.add_attribute(attr)

    check_gold(mod, "test_simple_pipeline", insert_pipeline_stages=True)


def test_auto_insert_clk_gate_skip():
    from _kratos.passes import auto_insert_clock_enable

    class Mod(Generator):
        def __init__(self, add_self, add_always):
            super(Mod, self).__init__("mod")
            clk = self.clock("clk")
            a = self.var("a", 1)
            self.clock_en("clk_en")

            @always_ff((posedge, clk))
            def code():
                a = 1

            b = self.add_always(code)
            if add_always:
                b.add_attribute("dont_touch")
            if add_self:
                self.add_attribute("dont_touch")
    m = Mod(False, False)
    auto_insert_clock_enable(m.internal_generator)
    src = verilog(m)["mod"]
    assert "if (clk_en)" in src
    m = Mod(False, True)
    auto_insert_clock_enable(m.internal_generator)
    src = verilog(m)["mod"]
    assert "if (clk_en)" not in src
    m = Mod(True, False)
    auto_insert_clock_enable(m.internal_generator)
    src = verilog(m)["mod"]
    assert "if (clk_en)" not in src


def test_gen_inst_lift(check_gold):
    num_inst = 4
    parent = Generator("parent")
    clk = parent.clock("clk")
    a_array = parent.var("a", 1, size=4)
    b_array = parent.var("b", 1, size=4)
    children = []

    for i in range(num_inst):
        child = Generator("child")
        child.clock("clk")
        a = child.input("a", 1)
        b = child.output("b", 1)
        child.wire(a, b)
        parent.add_child(f"child_{i}", child,
                         clk=clk,
                         a=a_array[i],
                         b=b_array[i])
        children.append(child)

    check_gold(parent, "test_gen_inst_lift", lift_genvar_instances=True)
    name = children[1].internal_generator.handle_name()
    assert name == "parent.child.inst[1]"

    # another one that will fail the genvar test
    Generator.clear_context()
    num_inst = 4
    parent = Generator("parent")
    clk = parent.clock("clk")
    a_array = parent.var("a", 1, size=2)

    for i in range(2):
        child = Generator("child")
        child.clock("clk")
        child.input("a", 1)
        parent.add_child(f"child_{i}", child,
                         clk=clk,
                         a=a_array[0])
    src = verilog(parent, lift_genvar_instances=True)["parent"]
    assert "genvar" not in src
